/* 
   A C-program for MT19937, with initialization improved 2002/1/26.
   Coded by Takuji Nishimura and Makoto Matsumoto.

   Before using, initialize the state by using init_genrand(seed)  
   or init_by_array(init_key, key_length).

   Copyright (C) 1997 - 2002, Makoto Matsumoto and Takuji Nishimura,
   All rights reserved.                          

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

     1. Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.

     2. Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.

     3. The names of its contributors may not be used to endorse or promote 
        products derived from this software without specific prior written 
        permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


   Any feedback is very welcome.
   http://www.math.keio.ac.jp/matumoto/emt.html
   email: matumoto@math.keio.ac.jp
*/

//================================================================================================================
// INCLUDE
//================================================================================================================
#include "LunaBase.h"

#include "LunaRand.h"

namespace lunadxmath{
//================================================================================================================
// CONST
//================================================================================================================
static const Uint32 N			= 624UL;
static const Uint32 M			= 397UL;
static const Uint32 MATRIX_A		= 0x9908B0DFUL;
static const Uint32 UPPER_MASK	= 0x80000000UL;
static const Uint32 LOWER_MASK	= 0x7FFFFFFFUL;


//================================================================================================================
// VARIABLE
//================================================================================================================
static Uint32 MersenneTwister[N];
static Sint32 MersenneTwisterCount = N + 1;


//================================================================================================================
/**
*/
//================================================================================================================
Uint32 LunaRand::GenerateInt32( void )
{
    Uint32 Mag01[2] = { 0x00000000UL, MATRIX_A };
    Uint32 Param;

    if ( MersenneTwisterCount >= N )
	{
        Sint32 kk = 0;

		while ( kk < N - M )
		{
            Param = (MersenneTwister[kk] & UPPER_MASK) | (MersenneTwister[kk+1] & LOWER_MASK);
            MersenneTwister[kk] = MersenneTwister[kk + M] ^ (Param >> 1) ^ Mag01[Param & 0x01];

			kk++;
		}

		while ( kk < N - 1 )
		{
            Param = (MersenneTwister[kk] & UPPER_MASK) | (MersenneTwister[kk + 1] & LOWER_MASK);
            MersenneTwister[kk] = MersenneTwister[kk + (M - N)] ^ (Param >> 1) ^ Mag01[Param & 0x00000001UL];

			kk++;
		}

		Param = (MersenneTwister[N - 1]&UPPER_MASK) | (MersenneTwister[0] & LOWER_MASK);
        MersenneTwister[N - 1] = MersenneTwister[M-1] ^ (Param >> 1) ^ Mag01[Param & 0x00000001UL];

        MersenneTwisterCount = 0;
    }
  
    Param = MersenneTwister[MersenneTwisterCount++];

    /* Tempering */
    Param ^= (Param >> 11);
    Param ^= (Param <<  7) & 0x9D2C5680UL;
    Param ^= (Param << 15) & 0xEFC60000UL;
    Param ^= (Param >> 18);

    return Param;
}

//================================================================================================================
/**
*/
//================================================================================================================
void LunaRand::Seed( Uint32 Param )
{
    MersenneTwister[0] = Param & 0xFFFFFFFFUL;
    for ( MersenneTwisterCount = 1; MersenneTwisterCount < N; MersenneTwisterCount++ )
	{
        MersenneTwister[MersenneTwisterCount] = (1812433253UL * (MersenneTwister[MersenneTwisterCount-1] ^ (MersenneTwister[MersenneTwisterCount-1] >> 30)) + MersenneTwisterCount); 
        MersenneTwister[MersenneTwisterCount] &= 0xFFFFFFFFUL;
    }
}

//================================================================================================================
/**
*/
//================================================================================================================
Uint32 LunaRand::GetInt32( void )
{
    return GenerateInt32();
}

//================================================================================================================
/**
*/
//================================================================================================================
#ifndef __GCC__
Uint64 LunaRand::GetInt64( void )
{
	Uint64 a = GenerateInt32();
	Uint64 b = GenerateInt32();

	return (a << 32) | b;
}
#endif // __GCC__

//================================================================================================================
/**
*/
//================================================================================================================
Float LunaRand::GetFloat32( void )
{
    return ((Float)GenerateInt32() + 0.5f) * (1.0f / 4294967296.0f); 
}

//================================================================================================================
/**
*/
//================================================================================================================
Float64 LunaRand::GetFloat53( void )
{
    Uint32 a = GenerateInt32() >> 5;
	Uint32 b = GenerateInt32() >> 6;

	return ((Float64)a * 67108864.0 + (Float64)b) * (1.0 / 9007199254740992.0);
}

//================================================================================================================
/**
*/
//================================================================================================================
Sint32 LunaRand::GetInt( Sint32 Min, Sint32 Max )
{
	Sint32 a = (Sint32)(GetInt32() >> 1);

	return Min + (a % (Max + 1 - Min));
}

//================================================================================================================
/**
*/
//================================================================================================================
Float LunaRand::GetFloat( Float Min, Float Max )
{
	Float a = GetFloat32();

	return Min + (a * (Max - Min));
}

}